from datetime import datetime
from functools import update_wrapper

from flask import request, Response

from xf_auth.Auth import Auth
from xf_auth.HTTPStatus import UNAUTHORIZED, SESSION_EXPIRED, FORBIDDEN


class StatefulSession:
	session: list = []
	session_lifetime: int = 6_000

	@staticmethod
	def new_session(data: dict) -> str or None:
		stored_token = StatefulSession.__get_token_from_data(data)
		if stored_token is not None:
			return stored_token

		token: str = StatefulSession.__new_token(data)
		StatefulSession.session.append({
			"token": token,
			"data": data,
			"session_start": datetime.now()
		})
		return token

	@staticmethod
	def __new_token(data: dict) -> str:
		if Auth.class_name is not None:
			aux_obj = Auth.class_name()
			for key, value in data.items():
				if key in aux_obj.__dict__:
					aux_obj.__setattr__(key, value)
				else:
					raise AttributeError("Attribute is not part of provided User origin class")
			return Auth.generate_token(aux_obj)
		else:
			raise TypeError("User origin is not set. Set it before setting user keys with Auth.set_user_origin")

	@staticmethod
	def get_data(token: str, stateless: bool = False) -> dict or None:
		data: dict or None = None
		for session in StatefulSession.session:
			if session["token"] == token:
				data = session["data"]
				break
		if data is None and stateless:
			data = Auth.decode_token(token)
		return data

	@staticmethod
	def get_session(token: str) -> dict or None:
		session: dict or None = None
		for saved_session in StatefulSession.session:
			if saved_session["token"] == token:
				session = saved_session
				break
		return session

	@staticmethod
	def delete_session(token: str) -> bool:
		deleted: bool = False
		stored_session: dict = StatefulSession.get_session(token)
		if stored_session is not None:
			StatefulSession.session.remove(stored_session)
			deleted = True
		return deleted

	@staticmethod
	def requires_token(operation):
		def verify_auth(*args, **kwargs):
			token = request.headers.get("token")
			if token is not None:
				session = StatefulSession.get_session(token)
				if session is not None:
					now = datetime.now()
					if (now - session["session_start"]).total_seconds() < StatefulSession.session_lifetime:
						session["session_start"] = now
						response = operation(*args, **kwargs)
					else:
						response = Response(status=SESSION_EXPIRED)
				else:
					response = Response(status=UNAUTHORIZED)
			else:
				response = Response(status=UNAUTHORIZED)
			return response

		return update_wrapper(verify_auth, operation)

	@staticmethod
	def requires_role(role: str):
		def decorator(operation):
			def verify_role(*args, **kwargs):
				if Auth.role_attribute is not None:
					token = request.headers.get("token")
					if token is not None:
						session = StatefulSession.get_session(token)
						if session is not None:
							now = datetime.now()
							if (now - session["session_start"]).total_seconds() < StatefulSession.session_lifetime:
								if str(session["data"][Auth.role_attribute]).lower() == role.lower():
									session["session_start"] = now
									response = operation(*args, **kwargs)
								else:
									response = Response(status=FORBIDDEN)
							else:
								response = Response(status=SESSION_EXPIRED)
						else:
							response = Response(status=SESSION_EXPIRED)
					else:
						response = Response(status=UNAUTHORIZED)
				else:
					raise TypeError(
						"User role attribute is not set. Set it with Auth.set_role_attribute")
				return response

			return update_wrapper(verify_role, operation)

		return decorator

	@staticmethod
	def __get_token_from_data(data: dict) -> str or None:
		token: str or None = None
		for session in StatefulSession.session:
			if session["data"]["email"] == data["email"] and session["data"]["password"] == data["password"]:
				token = session["token"]
		return token

	@staticmethod
	def __is_session_in(data: dict) -> bool:
		is_in: bool = False
		if data is not None:
			for session in StatefulSession.session:
				if session["data"]["email"] == data["email"] and session["data"]["password"] == data["password"]:
					is_in = True
					break
		return is_in



from datetime import datetime
from functools import update_wrapper
from typing import Type

from cryptography.fernet import Fernet
from flask import Response, request, session
from werkzeug.security import generate_password_hash, check_password_hash

from xf_auth.HTTPStatus import UNAUTHORIZED, SESSION_EXPIRED, FORBIDDEN, NOT_ACCEPTABLE
from xf_auth.Util import decode, included, encode


class Auth:
	secret_password: bytes = None
	class_name: Type[object] = None
	user_keys: list = None
	role_attribute: str = None

	@staticmethod
	def set_password():
		Auth.secret_password = Fernet.generate_key()

	@staticmethod
	def set_user_origin(class_name: Type[object]) -> None:
		Auth.class_name = class_name

	@staticmethod
	def set_user_keys(keys: list) -> None:
		if Auth.class_name is not None:
			aux = Auth.class_name()
			keys.sort()
			for key in keys:
				try:
					aux.__getattribute__(key)
				except AttributeError as exception:
					raise Exception(
						"Attribute not in instance of class. Attributes should initialize on init") from exception
			Auth.user_keys = keys
		else:
			raise TypeError("User origin is not set. Set it before setting user keys with Auth.set_user_origin")

	@staticmethod
	def set_role_attribute(attribute_name: str) -> None:
		if Auth.class_name is not None:
			aux = Auth.class_name()
			try:
				aux.__getattribute__(attribute_name)
			except AttributeError as exception:
				raise Exception(
					"Attribute is not in instance of class. Attributes should initialize on init") from exception
			Auth.role_attribute = attribute_name
		else:
			raise TypeError("User origin is not set. Set it before setting role attribute with Auth.set_user_origin")

	@staticmethod
	def requires_stateless_token(operation):
		def verify_auth(*args, **kwargs):
			token = request.headers.get("token")
			saved_token = None
			try:
				saved_token = session["token"]
				response = Response(status=UNAUTHORIZED)
				if token is not None and saved_token is not None and token == saved_token:
					session.modified = True
					response = operation(*args, **kwargs)
			except KeyError:
				response = Response(status=UNAUTHORIZED)
				if token is not None and saved_token is None:
					response = Response(status=SESSION_EXPIRED)
			return response

		return update_wrapper(verify_auth, operation)

	@staticmethod
	def requires_role(role: str):
		def decorator(operation):
			def verify_role(*args, **kwargs):
				token = request.headers.get("token")
				response = Response(status=FORBIDDEN)
				if token is not None:
					values = Auth.decode_token(token)
					response = Response(status=FORBIDDEN)
					if str(values["is_owner"]) == str(role):
						response = operation(*args, **kwargs)
				return response

			return update_wrapper(verify_role, operation)

		return decorator

	@staticmethod
	def requires_payload(required_fields: set):
		def decorator(operation):
			def verify_payload(*args, **kwargs):
				if not included(required_fields, request.json):
					response = Response(status=NOT_ACCEPTABLE)
				else:
					response = operation(*args, **kwargs)
				return response

			return update_wrapper(verify_payload, operation)

		return decorator

	@staticmethod
	def generate_token(user) -> str or None:
		response: str or None
		if Auth.class_name is not None:
			if isinstance(user, Auth.class_name):
				if Auth.secret_password is None:
					Auth.set_password()
				timestamp = datetime.now().strftime("%H:%M:%S")

				value: str = ""
				for key in Auth.user_keys:
					value += user.__getattribute__(key) + "/"

				if Auth.role_attribute is not None:
					current_role = user.__getattribute__(Auth.role_attribute)
					if current_role is not None:
						value += current_role + "/"
					else:
						value += "NONE/"

				value += timestamp
				response = encode(value, Auth.secret_password)
			else:
				raise TypeError("Provided user is not instance of User origin")
		else:
			raise TypeError("User origin is not set. Set it before setting role attribute with Auth.set_user_origin")
		return response

	@staticmethod
	def decode_token(token: str) -> dict:
		if Auth.secret_password is not None:
			decoded_token = decode(token, Auth.secret_password)
			decoded_token = decoded_token.split("/")
			aux_dict: dict = {}

			for key, value in zip(Auth.user_keys, decoded_token):
				aux_dict[key] = value

			if Auth.role_attribute is not None:
				aux_dict[Auth.role_attribute] = decoded_token[-2]
		else:
			raise TypeError("No token has been created. Cannot decode token")
		return aux_dict

	@staticmethod
	def crypt(string: str) -> str:
		return generate_password_hash(string)

	@staticmethod
	def check_crypted(crypted: str, clean: str) -> str:
		return check_password_hash(crypted, clean)



OK = 200
RESOURCE_CREATED = 201
ACCEPTED = 202
NO_CONTENT = 204
RESET_CONTENT = 205
PARTIAL_CONTENT = 206
FILE_UPLOADED = 226
MOVED_PERMANENTLY = 301
BAD_REQUEST = 400
UNAUTHORIZED = 401
FORBIDDEN = 403
NOT_FOUND = 404
METHOD_NOT_ALLOWED = 405
NOT_ACCEPTABLE = 406
CONFLICT = 409
SESSION_EXPIRED = 419
UNSUPPORTED_MEDIA_TYPE = 4015
INTERNAL_SERVER_ERROR = 500
NOT_IMPLEMENTED = 501



from setuptools import setup

setup(
	name='xf_auth',
	version='0.5.3',
	description='A package to handle auth on flask. Made by Xatal',
	url='https://github.com/Masakuata/xatal_flask_auth',
	author='Edson Manuel Carballo Vera',
	author_email='edsonmanuelcarballovera@gmail.com',
	license='MIT License',
	packages=['xf_auth'],
	install_requires=['cryptography==3.1.1', 'Flask'],

	classifiers=[
		'Development Status :: 2 - Pre-Alpha',
		'Framework :: Flask',
		'Intended Audience :: Developers',
		'License :: OSI Approved :: MIT License',
		'Operating System :: POSIX :: Linux',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3.9',
		'Programming Language :: Python :: 3.10',
		'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
		'Topic :: Internet :: WWW/HTTP :: Session'
	],
)
